"""This module defines a pipeline to create RSS feeds.

The module takes PodcastEpisodeItems and PodcastDataItems and exports them using
PodcastToS3ItemExporter or PodcastToFileItemExporter depending on the URI defined.
Remember that to use PodcastToS3ItemExporter you need to have boto3 installed.
"""
from urllib.parse import urlparse

from scrapy_podcast_rss.exceptions import InvalidItemException
from scrapy_podcast_rss.exporters import PodcastToS3ItemExporter, PodcastToFileItemExporter
from scrapy_podcast_rss.items import PodcastEpisodeItem, PodcastDataItem


class PodcastPipeline:
    """Pipeline that takes podcast items to the exporter.

    This class takes PodcastEpisodeItems and at one PodcastDataItem,
    and depending on the URI defined in the settings, exports them using
    either PodcastToS3ItemExporter or PodcastToFileItemExporter.
    """

    def __init__(self):
        """Initializes the pipeline.
        It creates an empty (_episodes_item_list) list to store
        PodcastEpisodeItems and a variable (_podcast_data_item) to store
        one PodcastDataItem.
        """
        self._episodes_item_list = []
        self._podcast_data_item = None

    def close_spider(self, spider):
        """Exports each item and closes the exporter.

        Args:
            spider: A spider.
        """
        exporter = self._get_exporter(spider)
        exporter.start_exporting()

        for episode_item in self._episodes_item_list:
            exporter.export_item(episode_item)

        exporter.finish_exporting()

    def _get_exporter(self, spider):
        """Initializes and returns a exporter.

        The exporter selected depends on the type of URI defined in the settings
        under OUTPUT_URI.

        Args:
            spider: A spider.

        Returns:
            PodcastToS3ItemExporter or PodcastToFileItemExporter
        """
        if self._podcast_data_item is None:
            raise ValueError("You need to yield at least one PodcastDataItem.")

        title = self._podcast_data_item.get('title')
        description = self._podcast_data_item.get('description')
        url = self._podcast_data_item.get('url')
        image_url = self._podcast_data_item.get('image_url')

        uri = spider.settings.get('OUTPUT_URI')
        if self._uri_for_s3(uri):
            exp = PodcastToS3ItemExporter(uri, title=title, description=description, url=url, image_url=image_url)
        else:
            exp = PodcastToFileItemExporter(uri, title=title, description=description, url=url, image_url=image_url)

        return exp

    def process_item(self, item, spider):
        """Processes each yielded item.

        This function only allows PodcastEpisodeItems and PodcastDataItems.
        It only stores the last PodcastDataItem received.

        Args:
            item: PodcastEpisodeItem or PodcastDataItem.
            spider: A spider.

        Returns:
            item: PodcastEpisodeItem or PodcastDataItem.
        """
        if isinstance(item, PodcastEpisodeItem):
            self._episodes_item_list.append(item)
        elif isinstance(item, PodcastDataItem):
            self._podcast_data_item = item
        else:
            raise InvalidItemException(f"Only PodcastEpisodeItem and PodcastDataItem are allowed, not {type(item)}")

        return item

    @staticmethod
    def _uri_for_s3(uri):
        """Checks if the URI corresponds to an S3 url."""
        url = urlparse(uri)
        return url.scheme == 's3'



"""This module defines the items that will contain the podcast information.
"""

import scrapy


class PodcastDataItem(scrapy.Item):
    """General information about the podcast."""
    title = scrapy.Field()
    description = scrapy.Field()
    url = scrapy.Field()
    image_url = scrapy.Field()


class PodcastEpisodeItem(scrapy.Item):
    """Information about each episode of the podcast."""
    title = scrapy.Field()
    description = scrapy.Field()
    publication_date = scrapy.Field()
    audio_url = scrapy.Field()
    guid = scrapy.Field()



__all__ = ['PodcastDataItem', 'PodcastEpisodeItem', 'PodcastPipeline']

from scrapy_podcast_rss.__version__ import __version__, __author__, __author_email__
from scrapy_podcast_rss.items import PodcastDataItem, PodcastEpisodeItem  # noqa
from scrapy_podcast_rss.pipelines import PodcastPipeline  # noqa



"""Custom exporters that generate and save RSS feeds.
"""
import abc
from urllib.parse import urlparse

from feedgen.feed import FeedGenerator
from scrapy.exporters import BaseItemExporter

try:
    import boto3
except ImportError:
    boto3 = None


class PodcastBaseItemExporter(BaseItemExporter, metaclass=abc.ABCMeta):
    """Item exporter base class designed to generate RSS feeds.

    The class uses feedgen to generate the RSS content.
    Subclasses are expected to implement the method save_to_storage.
    """

    def __init__(self, uri, title, description, url, image_url, **kwargs):
        """Initializes the exporter.

        Args:
            uri: Where to save the feed.
            title: Podcast title.
            description: Description of the podcast.
            url: Url of the podcast.
            image_url: Main image of the podcast.
            **kwargs: Any extra argument for BaseItemExporter.
        """
        super().__init__(**kwargs)
        self.uri = uri

        self.fg = FeedGenerator()
        self.fg.load_extension('podcast')

        self.fg.title(title)
        self.fg.description(description)
        self.fg.link(href=url)
        self.fg.image(image_url)
        self.fg._FeedGenerator__rss_lastBuildDate = None  # This prevents Plex from confusing pubDate with lastBuildDate

    def export_item(self, item):
        """Adds a new entry to the RSS feed.

        Args:
            item: A PodcastEpisodeItem.
        """
        fe = self.fg.add_entry()

        title = item.get('title')
        description = item.get('description')
        publication_date = item.get('publication_date')
        audio_url = item.get('audio_url')
        guid = item.get('guid')

        fe.title(title)
        fe.description(description)
        fe.published(publication_date)
        fe.enclosure(audio_url, 0, 'audio/mpeg')
        fe.guid(guid)

    def finish_exporting(self):
        """Generates the RSS content and saves it to a file"""
        rss_content = self.fg.rss_str(pretty=True)
        self.save_to_storage(rss_content)

    @abc.abstractmethod
    def save_to_storage(self, rss_content):
        """Subclasses must implement a way of saving the content.
        """
        pass


class PodcastToFileItemExporter(PodcastBaseItemExporter):
    """Exporter that saves the RSS feed to a local file."""

    def save_to_storage(self, content):
        """Saves the content to a local file."""
        f = open(self.uri, 'wb')
        f.write(content)
        f.close()


class PodcastToS3ItemExporter(PodcastBaseItemExporter):
    """Exporter that saves the RSS feed to an S3 bucket

    NOTE 1: This exporter needs boto3 in order to operate.
    NOTE 2: By default, the file is configured as PUBLIC-READ.
    """
    def __init__(self, *args, **kwargs):
        """Checks if boto3 is installed."""
        super().__init__(*args, **kwargs)
        if boto3 is None:
            raise ImportError("You need to install boto3 to export items to S3. More information here: "
                              "https://boto3.amazonaws.com/v1/documentation/api/latest/guide/quickstart.html")

    def save_to_storage(self, content):
        """Saves the content with PUBLIC-READ permission to an S3 bucket.

        The uri defined in the settings under the name OUTPUT_URI needs
        to have the following format:
            s3://{bucket-name}/{rss-feed-filename}.xml
        (the extension can also be changed.)
        """
        url = urlparse(self.uri)
        bucket = url.netloc
        filename = url.path.split("/")[-1]

        s3 = boto3.resource('s3')
        s3_object = s3.Object(bucket, filename)
        s3_object.put(Body=content, ACL='public-read')



import os

import setuptools

here = os.path.abspath(os.path.dirname(__file__))


def read(file_name):
    return open(os.path.join(here, file_name)).read()


about = {}
with open(os.path.join(here, 'scrapy_podcast_rss', '__version__.py'), 'r') as f:
    exec(f.read(), about)

setuptools.setup(
    name="scrapy-podcast-rss",
    version=about["__version__"],
    author=about["__author__"],
    author_email=about["__author_email__"],
    description=about["__description__"],
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    url="https://github.com/igarizio/scrapy-podcast-rss",
    packages=['scrapy_podcast_rss'],
    install_requires=["Scrapy>=1.7.3", "feedgen==0.9.0"],
    extras_require={
        's3_storage': ["boto3"],
        'tests': ["pytest==5.2.4"]
    },
    python_require='>=3.6',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
